"""
Buildtool independent utils to discover plugins from the codebase.
"""
import inspect
import logging
from types import ModuleType
import typing as t

from plux import PluginFinder, PluginSpecResolver, PluginSpec

LOG = logging.getLogger(__name__)


class ModuleScanningPluginFinder(PluginFinder):
    """
    A PluginFinder that scans the members of given modules for available PluginSpecs. Each member is evaluated with a
    PluginSpecResolver, and all successful calls resulting in a PluginSpec are collected and returned. This is used
    at build time to scan all modules that the build tool finds, and is an expensive solution that would not be
    practical at runtime.
    """

    def __init__(self, modules: t.Iterable[ModuleType], resolver: PluginSpecResolver = None) -> None:
        super().__init__()
        self.modules = modules
        self.resolver = resolver or PluginSpecResolver()

    def find_plugins(self) -> t.List[PluginSpec]:
        plugins = list()

        for module in self.modules:
            LOG.debug("scanning module %s", module.__name__)
            members = inspect.getmembers(module)

            for member in members:
                if type(member) is tuple:
                    try:
                        spec = self.resolver.resolve(member[1])
                        plugins.append(spec)
                        LOG.debug("found plugin spec in %s:%s %s", module.__name__, member[0], spec)
                    except Exception:
                        pass

        return plugins



import abc
import functools
import inspect
import typing as t


class PluginException(Exception):
    def __init__(self, message, namespace: str = None, name: str = None) -> None:
        super().__init__(message)
        self.namespace = namespace
        self.name = name


class PluginDisabled(PluginException):
    def __init__(self, namespace: str, name: str, reason: str = None):
        message = f"plugin {namespace}:{name} is disabled"
        if reason:
            message = f"{message}, reason: {reason}"
        super(PluginDisabled, self).__init__(message)
        self.namespace = namespace
        self.name = name


class Plugin(abc.ABC):
    """A generic LocalStack plugin.

    A Plugin's purpose is to be loaded dynamically at runtime and defer code imports into the Plugin::load method.
    Abstract subtypes of plugins (e.g., a LocalstackCliPlugin) may overwrite the load method with concrete call
    arguments that they agree upon with the PluginManager. In other words, a Plugin and a PluginManager for that
    particular Plugin have an informal contracts to use the same argument types when load is invoked.
    """

    namespace: str
    name: str
    requirements: t.List[str]

    def should_load(self) -> bool:
        return True

    def load(self, *args, **kwargs):
        """
        Called by a PluginLoader when it loads the Plugin.
        """
        return None


PluginType = t.Type[Plugin]
PluginFactory = t.Callable[[], Plugin]


class PluginSpec:
    """
    A PluginSpec describes a plugin through a namespace and its unique name within in that namespace, and holds the
    imported code that can instantiate the plugin (a PluginFactory). In the simplest case, the PluginFactory that can
    just be the Plugin's class.

    Internally a PluginSpec is essentially a wrapper around an importlib EntryPoint. An entrypoint is a tuple: (
    "name", "module:object") inside a namespace that can be loaded. The entrypoint object of a Plugin can point to a
    PluginSpec, or a Plugin that defines its own namespace and name, in which case the PluginSpec will be instantiated
    dynamically by, e.g., a PluginSpecResolver.
    """

    namespace: str
    name: str
    factory: PluginFactory

    def __init__(
        self,
        namespace: str,
        name: str,
        factory: PluginFactory,
    ) -> None:
        super().__init__()
        self.namespace = namespace
        self.name = name
        self.factory = factory

    def __str__(self):
        return "PluginSpec(%s.%s = %s)" % (self.namespace, self.name, self.factory)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.namespace == other.namespace
            and self.name == other.name
            and self.factory == other.factory
        )


class PluginFinder(abc.ABC):
    """
    Basic abstractions to find plugins, either at build time (e.g., using the PackagePathPluginFinder) or at run time
    (e.g., using StevedorePluginFinder that finds plugins from entrypoints)
    """

    def find_plugins(self) -> t.List[PluginSpec]:
        raise NotImplementedError  # pragma: no cover


class PluginSpecResolver:
    """
    A PluginSpecResolver finds or creates PluginSpec instances from sources, e.g., from analyzing a Plugin class.
    """

    def resolve(self, source: t.Any) -> PluginSpec:
        """
        Tries to create a PluginSpec from the given source.

        :param source: anything that can produce a PluginSpec (Plugin class, ...)
        :return: a PluginSpec instance
        """
        if isinstance(source, PluginSpec):
            return source

        if inspect.isclass(source):
            if issubclass(source, Plugin):
                return PluginSpec(source.namespace, source.name, source)

        if inspect.isfunction(source):
            spec = getattr(source, "__pluginspec__", None)
            if spec and isinstance(spec, PluginSpec):
                return spec

        # TODO: add more options to specify plugin specs

        raise ValueError("cannot resolve plugin specification from %s" % source)


class PluginLifecycleListener:  # pragma: no cover
    """
    Listener that can be attached to a PluginManager to react to plugin lifecycle events.
    """

    def on_resolve_exception(self, namespace: str, entrypoint, exception: Exception):
        pass

    def on_resolve_after(self, plugin_spec: PluginSpec):
        pass

    def on_init_exception(self, plugin_spec: PluginSpec, exception: Exception):
        pass

    def on_init_after(self, plugin_spec: PluginSpec, plugin: Plugin):
        pass

    def on_load_before(
        self,
        plugin_spec: PluginSpec,
        plugin: Plugin,
        load_args: t.Union[t.List, t.Tuple],
        load_kwargs: t.Dict,
    ):
        pass

    def on_load_after(self, plugin_spec: PluginSpec, plugin: Plugin, load_result: t.Any = None):
        pass

    def on_load_exception(self, plugin_spec: PluginSpec, plugin: Plugin, exception: Exception):
        pass


class CompositePluginLifecycleListener(PluginLifecycleListener):
    """A PluginLifecycleListener decorator that dispatches to multiple delegates."""

    listeners: t.List[PluginLifecycleListener]

    def __init__(self, initial: t.Iterable[PluginLifecycleListener] = None):
        self.listeners = list(initial) if initial else []

    def add_listener(self, listener: PluginLifecycleListener):
        self.listeners.append(listener)

    def on_resolve_exception(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_resolve_exception(*args, **kwargs)

    def on_resolve_after(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_resolve_after(*args, **kwargs)

    def on_init_exception(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_init_exception(*args, **kwargs)

    def on_init_after(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_init_after(*args, **kwargs)

    def on_load_before(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_load_before(*args, **kwargs)

    def on_load_after(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_load_after(*args, **kwargs)

    def on_load_exception(self, *args, **kwargs):
        for listener in self.listeners:
            listener.on_load_exception(*args, **kwargs)


class FunctionPlugin(Plugin):
    """
    Exposes a function as a Plugin.
    """

    fn: t.Callable

    def __init__(
        self,
        fn: t.Callable,
        should_load: t.Union[bool, t.Callable[[], bool]] = None,
        load: t.Callable = None,
    ) -> None:
        super().__init__()
        self.fn = fn
        self._should_load = should_load
        self._load = load

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def load(self, *args, **kwargs):
        if self._load:
            return self._load(*args, **kwargs)

    def should_load(self) -> bool:
        if self._should_load is not None:
            if type(self._should_load) == bool:
                return self._should_load
            else:
                return self._should_load()

        return True


def plugin(
    namespace,
    name=None,
    should_load: t.Union[bool, t.Callable[[], bool]] = None,
    load: t.Callable = None,
):
    """
    Expose a function as discoverable and loadable FunctionPlugin.

    :param namespace: the plugin namespace
    :param name: the name of the plugin (by default the function name will be used)
    :param should_load: optional either a boolean value or a callable returning a boolean
    :param load: optional load function
    :return: plugin decorator
    """

    def wrapper(fn):
        plugin_name = name or fn.__name__

        # this causes the plugin framework to point the entrypoint to the original function rather than the
        # nested factory function (which would not be resolvable)
        @functools.wraps(fn)
        def factory():
            fn_plugin = FunctionPlugin(fn, should_load=should_load, load=load)
            fn_plugin.namespace = namespace
            fn_plugin.name = plugin_name
            return fn_plugin

        # at discovery-time the factory will point to the method being decorated, and at load-time the factory from
        # this spec instance be used instead of the one being created
        fn.__pluginspec__ = PluginSpec(namespace, plugin_name, factory)

        return fn

    return wrapper



from plux.build.setuptools import (
    entry_points_from_egg_info,
    get_plux_json_path,
    load_plux_entrypoints,
    plugins,
    update_entrypoints,
)

__all__ = [
    "plugins",
    "load_plux_entrypoints",
    "get_plux_json_path",
    "update_entrypoints",
    "entry_points_from_egg_info",
]



"""
Bindings to integrate plux into setuptools build processes.
"""
import importlib
import json
import logging
import os
import re
import sys
import typing as t

import setuptools
from setuptools.command.egg_info import InfoCommon, write_entries

from plux.core.entrypoint import EntryPointDict, to_entry_point_dict, discover_entry_points
from plux.core.plugin import PluginFinder, PluginSpec
from plux.runtime.metadata import Distribution  # FIXME
from .discovery import ModuleScanningPluginFinder

LOG = logging.getLogger(__name__)


class plugins(InfoCommon, setuptools.Command):
    description = "Discover plux plugins and store them in .egg_info"

    user_options = [
        # TODO
    ]

    egg_info: str

    def initialize_options(self) -> None:
        self.plux_json_path = None

    def finalize_options(self) -> None:
        self.plux_json_path = get_plux_json_path(self.distribution)

    def run(self) -> None:
        plugin_finder = PluginFromPackageFinder(DistributionPackageFinder(self.distribution))
        ep = discover_entry_points(plugin_finder)

        self.debug_print(f"writing discovered plugins into {self.plux_json_path}")
        self.mkpath(os.path.dirname(self.plux_json_path))
        with open(self.plux_json_path, "w") as fd:
            json.dump(ep, fd)

        update_entrypoints(self.distribution, ep)


def find_plugins(where=".", exclude=(), include=("*",)) -> EntryPointDict:
    """
    Utility for setup.py that collects all plugins from the specified path, and creates a dictionary for entry_points.

    For example:

    setup(
        entry_points=find_plugins()
    )
    """

    return discover_entry_points(
        PackagePathPluginFinder(where=where, exclude=exclude, include=include)
    )


def load_plux_entrypoints(cmd, file_name, file_path):
    if not os.path.exists(file_path):
        return

    cmd.debug_print(f"extend entrypoints with plux plugins from {file_path}")
    with open(file_path, "r") as fd:
        ep = json.load(fd)

    update_entrypoints(cmd.distribution, ep)

    # this is kind of a hack, but we cannot rely on load_plux_entrypoints being called before the regular
    # entry_points.txt egg_info.writers plugin to write the updated entry points
    ep_file = "entry_points.txt"
    ep_path = os.path.join(os.path.dirname(file_path), ep_file)
    write_entries(cmd, ep_file, ep_path)


def get_plux_json_path(distribution):
    dirs = distribution.package_dir
    egg_base = (dirs or {}).get("", os.curdir)
    egg_info_dir = _to_filename(_safe_name(distribution.get_name())) + ".egg-info"
    egg_info_dir = os.path.join(egg_base, egg_info_dir)
    return os.path.join(egg_info_dir, "plux.json")


def update_entrypoints(distribution, ep: EntryPointDict):
    if distribution.entry_points is None:
        distribution.entry_points = {}

    # TODO: merge entry point groups
    distribution.entry_points.update(ep)


def load_entry_points(
    where=".", exclude=(), include=("*",), merge: EntryPointDict = None
) -> EntryPointDict:
    """
    Finds plugins and builds and entry point map. This is to be used in a setup.py in the setup call:

    setup(
        entry_points=load_entry_points(exclude=("tests", "tests.*"), merge={
            "my.static.entrypoint": [
                "foo=bar"
            ]
        })
        ...
    )

    This is a hack for installing from source distributions. When running pip install on a source distribution,
    the egg_info directory is always re-built, even though it comes with the source distribution package data. This
    also means the entry points are resolved, and in extent, `find_plugins` is called, which is problematic at this
    point, because find_plugins will scan the code, and that will fail if requirements aren't yet installed,
    which they aren't when running pip install. However, since source distributions package the .egg-info directory,
    we can read the entry points from there instead, acting as sort of a cache.

    :param where: the file path to look for plugins (default, the current working dir)
    :param exclude: the glob patterns to exclude
    :param include: the glob patterns to include
    :param merge: a map of entry points that are always added
    """
    should_read, meta_dir = _should_read_existing_egg_info()
    if should_read:
        print("reading entry points from existing meta dir", meta_dir)
        eps = entry_points_from_egg_info(meta_dir)
    else:
        print("resolving plugin entry points")
        eps = find_plugins(where, exclude, include)

    if merge:
        # TODO: merge the dicts instead of overwriting (will involve de-duplicating entry points)
        eps.update(merge)

    return eps


def entry_points_from_egg_info(egg_info_dir: str) -> EntryPointDict:
    """
    Reads the entry_points.txt from a distribution meta dir (e.g., the .egg-info directory).
    """
    dist = Distribution.at(egg_info_dir)
    return to_entry_point_dict(dist.entry_points)


def _has_entry_points_cache() -> bool:
    egg_info_dir = _get_egg_info_dir()
    if not egg_info_dir:
        return False

    if not os.path.isfile(os.path.join(egg_info_dir, "entry_points.txt")):
        return False

    return True


def _should_read_existing_egg_info() -> t.Tuple[bool, t.Optional[str]]:
    # we want to read the .egg-info dir only if it exists, and if we are creating the egg_info or installing it with
    # pip install -e (which calls 'setup.py develop')

    if not (_is_pip_build_context() or _is_local_build_context()):
        return False, None

    egg_info_dir = _get_egg_info_dir()
    if not egg_info_dir:
        return False, None

    if not os.path.isfile(os.path.join(egg_info_dir, "entry_points.txt")):
        return False, None

    return True, egg_info_dir


def _is_pip_build_context():
    # when pip builds packages or wheels from source distributions, it creates a temporary directory with a marker
    # file that we can use to determine whether we are in such a build context.
    for f in os.listdir(os.getcwd()):
        if f == "pip-delete-this-directory.txt":
            return True

    return False


def _is_local_build_context():
    # when installing with `pip install -e` pip also tries to create a "modern-metadata" (dist-info) directory.
    if "dist_info" in sys.argv:
        try:
            i = sys.argv.index("--egg-base")
        except ValueError:
            return False

        if "pip-modern-metadata" in sys.argv[i + 1]:
            return True

    # it's unfortunately not really distinguishable whether or not a user calls `python setup.py develop` in the
    # project, or calls `pip install -e ..` to install the project from somewhere else.
    if len(sys.argv) > 1 and sys.argv[1] in ["egg_info", "develop"]:
        return True

    return False


def _get_egg_info_dir() -> t.Optional[str]:
    """
    Heuristic to find the .egg-info dir of the current build context.
    """
    cwd = os.getcwd()
    for d in os.listdir(cwd):
        if d.endswith(".egg-info"):
            return os.path.join(cwd, d)

    return None


def get_distribution_from_workdir(workdir: str) -> setuptools.Distribution:
    """
    Reads from the current workdir the available project configs and parses them to create a
    ``setuptools.Distribution`` object, which can later be used to invoke distutils commands.

    :param workdir: the workdir containing the project files
    :return: a distribution object
    """
    config_files = ["pyproject.toml", "setup.cfg"]
    config_files = [os.path.join(workdir, file) for file in config_files]
    config_files = [file for file in config_files if os.path.exists(file)]

    if not config_files:
        raise ValueError(f"no distribution config files found in {workdir}")

    dist = setuptools.Distribution()
    dist.parse_config_files(config_files)
    if os.path.exists("setup.py"):
        # use setup.py script if available
        dist.script_name = os.path.join(workdir, "setup.py")
    else:
        # else use a bogus file (seems to work regardless)
        dist.script_name = config_files[0]

    return dist


def _safe_name(name):
    """Convert an arbitrary string to a standard distribution name. Copied from pkg_resources.

    Any runs of non-alphanumeric/. characters are replaced with a single '-'.
    """
    return re.sub("[^A-Za-z0-9.]+", "-", name)


def _to_filename(name):
    """Convert a project or version name to its filename-escaped form. Copied from pkg_resources.

    Any '-' characters are currently replaced with '_'.
    """
    return name.replace("-", "_")


class _PackageFinder:
    """
    Generate a list of Python packages. How these are generated depends on the implementation.
    """

    def find_packages(self) -> t.Iterable[str]:
        raise NotImplementedError

    @property
    def path(self) -> str:
        raise NotImplementedError


class DistributionPackageFinder(_PackageFinder):
    """
    PackageFinder that returns the packages found in the distribution. The Distribution will already have a
    list of resolved packages depending on the setup config. For example, if a ``pyproject.toml`` is used,
    then the ``[tool.setuptools.package.find]`` config will be interpreted, resolved, and then
    ``distribution.packages`` will contain the resolved packages. This already contains namespace packages
    correctly if configured.
    """

    def __init__(self, distribution: Distribution):
        self.distribution = distribution

    def find_packages(self) -> t.Iterable[str]:
        return self.distribution.packages

    @property
    def path(self) -> str:
        if not self.distribution.package_dir:
            where = "."
        else:
            if self.distribution.package_dir[""]:
                where = self.distribution.package_dir[""]
            else:
                LOG.warning("plux doesn't know how to resolve multiple package_dir directories")
                where = "."
        return where


class DefaultPackageFinder(_PackageFinder):
    def __init__(self, where=".", exclude=(), include=("*",), namespace=True) -> None:
        self.where = where
        self.exclude = exclude
        self.include = include
        self.namespace = namespace

    def find_packages(self) -> t.Iterable[str]:
        if self.namespace:
            return setuptools.find_namespace_packages(self.where, self.exclude, self.include)
        else:
            return setuptools.find_packages(self.where, self.exclude, self.include)

    @property
    def path(self) -> str:
        return self.where


class PluginFromPackageFinder(PluginFinder):
    finder: _PackageFinder

    def __init__(self, finder: _PackageFinder):
        self.finder = finder

    def find_plugins(self) -> t.List[PluginSpec]:
        collector = ModuleScanningPluginFinder(self.load_modules())
        return collector.find_plugins()

    def load_modules(self):
        for module_name in self.list_module_names():
            try:
                yield importlib.import_module(module_name)
            except Exception as e:
                LOG.error("error importing module %s: %s", module_name, e)

    def list_module_names(self):
        # adapted from https://stackoverflow.com/a/54323162/804840
        from pkgutil import iter_modules

        modules = set()

        for pkg in self.finder.find_packages():
            modules.add(pkg)
            pkgpath = self.finder.path + os.sep + pkg.replace(".", os.sep)
            for info in iter_modules([pkgpath]):
                if not info.ispkg:
                    modules.add(pkg + "." + info.name)

        return modules


class PackagePathPluginFinder(PluginFromPackageFinder):
    """
    Uses setuptools and pkgutil to find and import modules within a given path and then uses a
    ModuleScanningPluginFinder to resolve the available plugins. The constructor has the same signature as
    setuptools.find_packages(where, exclude, include).
    """

    def __init__(self, where=".", exclude=(), include=("*",), namespace=True) -> None:
        super().__init__(DefaultPackageFinder(where, exclude, include, namespace=namespace))



"""
A plux frontend, currently only supports setuptools.
"""

import argparse
import json
import logging
import os

from plux.build.setuptools import (
    _get_egg_info_dir,
    find_plugins,
    get_distribution_from_workdir,
    get_plux_json_path,
)


def entrypoints(args):
    dist = get_distribution_from_workdir(args.workdir)

    print("discovering plugins ...")
    dist.run_command("plugins")

    print(f"building {dist.get_name().replace('-', '_')}.egg-info...")
    dist.run_command("egg_info")

    print("discovered plugins:")
    # discover plux plugins
    with open(get_plux_json_path(dist)) as fd:
        plux_json = json.load(fd)
    _pprint_plux_json(plux_json)


def discover(args):
    ep = find_plugins(exclude=("tests", "tests.*"))  # TODO: options
    _pprint_plux_json(ep)


def show(args):
    egg_info_dir = _get_egg_info_dir()
    if not egg_info_dir:
        print("no *.egg-info directory")
        return

    txt = os.path.join(egg_info_dir, "entry_points.txt")
    if not os.path.isfile(txt):
        print("no entry points to show")
        return

    with open(txt) as fd:
        print(fd.read())


def _pprint_plux_json(plux_json):
    print(json.dumps(plux_json, indent=2))


def main(argv=None):
    parser = argparse.ArgumentParser(description="Plux CLI frontend")
    parser.add_argument(
        "--workdir",
        type=str,
        default=os.getcwd(),
        help="overwrite the working directory",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="enable verbose logging")
    subparsers = parser.add_subparsers(title="commands", dest="command", help="Available commands")

    # Subparser for the 'generate' subcommand
    generate_parser = subparsers.add_parser(
        "entrypoints", help="Discover plugins and generate entry points"
    )
    generate_parser.set_defaults(func=entrypoints)

    # Subparser for the 'discover' subcommand
    discover_parser = subparsers.add_parser("discover", help="Discover plugins and print them")
    discover_parser.set_defaults(func=discover)

    # Subparser for the 'discover' subcommand
    show_parser = subparsers.add_parser("show", help="Show entrypoints that were generated")
    show_parser.set_defaults(func=show)

    args = parser.parse_args(argv)

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    os.chdir(args.workdir)

    if not hasattr(args, "func"):
        parser.print_help()
        exit(1)
    args.func(args)


if __name__ == "__main__":
    main()



from plux import __version__ as _version
from plux.core.plugin import (
    FunctionPlugin,
    Plugin,
    PluginDisabled,
    PluginException,
    PluginFinder,
    PluginLifecycleListener,
    PluginSpec,
    PluginSpecResolver,
    PluginType,
    plugin,
)
from plux.runtime.manager import PluginManager

name = "plugin"

__version__ = _version

__all__ = [
    "__version__",
    "FunctionPlugin",
    "Plugin",
    "PluginSpec",
    "PluginType",
    "PluginLifecycleListener",
    "PluginFinder",
    "PluginManager",
    "PluginSpecResolver",
    "PluginException",
    "PluginDisabled",
    "plugin",
]



"""
Module to provide easier high-level access to ``importlib.metadata`` in the context of plugins.
"""
import collections
import inspect
import sys
import typing as t
from functools import lru_cache
from importlib import metadata

if sys.version_info >= (3, 10):
    from importlib.metadata import packages_distributions as metadata_packages_distributions
else:

    def metadata_packages_distributions() -> t.Mapping[str, t.List[str]]:
        """Copied from Pyton 3.10 stdlib."""
        pkg_to_dist = collections.defaultdict(list)
        for dist in metadata.distributions():
            for pkg in (dist.read_text("top_level.txt") or "").split():
                pkg_to_dist[pkg].append(dist.metadata["Name"])
        return dict(pkg_to_dist)


from plux.core.plugin import PluginSpec

Distribution = metadata.Distribution


@lru_cache()
def packages_distributions() -> t.Mapping[str, t.List[str]]:
    """
    Cache wrapper around metadata.packages_distributions, which returns a mapping of top-level packages to
    their distributions.

    :return: package to distribution mapping
    """
    return metadata_packages_distributions()


def resolve_distribution_information(plugin_spec: PluginSpec) -> t.Optional[Distribution]:
    """
    Resolves for a PluginSpec the python distribution package it comes from. Currently, this raises an
    error for plugins that come from a namespace package (i.e., when a package is part of multiple
    distributions).

    :param plugin_spec: the plugin spec to resolve
    :return: the Distribution metadata if it exists
    """
    package = inspect.getmodule(plugin_spec.factory).__name__
    root_package = package.split(".")[0]
    distributions = packages_distributions().get(root_package)
    if not distributions:
        return None
    if len(distributions) > 1:
        raise ValueError("cannot deal with plugins that are part of namespace packages")

    return metadata.distribution(distributions[0])



import inspect
import typing as t
from collections import defaultdict

from .plugin import PluginFinder, PluginSpec


class EntryPoint(t.NamedTuple):
    name: str
    value: str
    group: str


EntryPointDict = t.Dict[str, t.List[str]]


def discover_entry_points(finder: PluginFinder) -> EntryPointDict:
    """
    Creates a dictionary for the entry_points attribute of setuptools' setup(), where keys are
    stevedore plugin namespaces, and values are lists of "name = module:object" pairs.

    :return: an entry_point dictionary
    """
    return to_entry_point_dict([spec_to_entry_point(spec) for spec in finder.find_plugins()])


def to_entry_point_dict(eps: t.List[EntryPoint]) -> EntryPointDict:
    result = defaultdict(list)
    names = defaultdict(set)  # book-keeping to check duplicates

    for ep in eps:
        if ep.name in names[ep.group]:
            raise ValueError("Duplicate entry point %s %s" % (ep.group, ep.name))

        result[ep.group].append("%s=%s" % (ep.name, ep.value))
        names[ep.group].add(ep.name)

    return result


def spec_to_entry_point(spec: PluginSpec) -> EntryPoint:
    module = inspect.getmodule(spec.factory).__name__
    name = spec.factory.__name__
    path = f"{module}:{name}"
    return EntryPoint(group=spec.namespace, name=spec.name, value=path)



from plux.runtime.metadata import (
    Distribution,
    metadata_packages_distributions,
    packages_distributions,
    resolve_distribution_information,
)

__all__ = [
    "resolve_distribution_information",
    "metadata_packages_distributions",
    "packages_distributions",
    "Distribution",
]



from plux.core.plugin import (
    FunctionPlugin,
    Plugin,
    PluginDisabled,
    PluginException,
    PluginFactory,
    PluginFinder,
    PluginLifecycleListener,
    PluginSpec,
    PluginSpecResolver,
    PluginType,
    plugin,
)

__all__ = [
    "FunctionPlugin",
    "Plugin",
    "PluginDisabled",
    "PluginException",
    "PluginFactory",
    "PluginFinder",
    "PluginLifecycleListener",
    "PluginSpec",
    "PluginSpecResolver",
    "PluginType",
    "plugin",
]



"""
Deprecated bindings, use plux imports instead, but you shouldn't use the internals in the first place.
"""
from plux.build.setuptools import find_plugins
from plux.core.entrypoint import (
    EntryPoint,
    EntryPointDict,
    spec_to_entry_point,
    to_entry_point_dict,
)

__all__ = [
    "find_plugins",
    "spec_to_entry_point",
    "EntryPointDict",
    "EntryPoint",
    "to_entry_point_dict",
]



"""Tools to resolve PluginSpec instances from entry points at runtime. Currently, stevedore does most of the heavy
lifting for us here (it caches entrypoints and can load them quickly)."""
import logging
import typing as t

from plux.core.plugin import PluginFinder, PluginSpec, PluginSpecResolver

LOG = logging.getLogger(__name__)


class StevedorePluginFinder(PluginFinder):
    """
    Uses a stevedore.Extension manager to resolve PluginSpec instances from entry points.
    """

    def __init__(
        self,
        namespace: str,
        on_resolve_exception_callback: t.Callable[[str, t.Any, Exception], None] = None,
        spec_resolver: PluginSpecResolver = None,
    ) -> None:
        super().__init__()
        self.namespace = namespace
        self.on_resolve_exception_callback = on_resolve_exception_callback
        self.spec_resolver = spec_resolver or PluginSpecResolver()

    def find_plugins(self) -> t.List[PluginSpec]:
        from stevedore import ExtensionManager
        from stevedore.exception import NoMatches

        manager = ExtensionManager(
            self.namespace,
            invoke_on_load=False,
            on_load_failure_callback=self._on_load_failure_callback,
        )

        # creates for each loadable stevedore extension a PluginSpec
        try:
            return manager.map(self.to_plugin_spec)
        except NoMatches:
            LOG.debug("no extensions found in namespace %s", self.namespace)
            return []

    def to_plugin_spec(self, ext) -> PluginSpec:
        """
        Convert a stevedore extension into a PluginSpec by using a spec_resolver.
        """
        try:
            return self.spec_resolver.resolve(ext.plugin)
        except Exception as e:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.exception(
                    "error resolving PluginSpec for plugin %s.%s", self.namespace, ext.name
                )

            self.on_resolve_exception_callback(self.namespace, ext.entry_point, e)

    def _on_load_failure_callback(self, _mgr, entrypoint, exception):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.error("error importing entrypoint %s: %s", entrypoint, exception)
        self.on_resolve_exception_callback(self.namespace, entrypoint, exception)



import logging
import threading
import typing as t

from plux.core.plugin import (
    Plugin,
    PluginDisabled,
    PluginException,
    PluginFinder,
    PluginLifecycleListener,
    PluginSpec,
)

from .metadata import Distribution, resolve_distribution_information
from .resolve import StevedorePluginFinder

LOG = logging.getLogger(__name__)

P = t.TypeVar("P", bound=Plugin)


def _call_safe(func: t.Callable, args: t.Tuple, exception_message: str):
    """
    Call the given function with the given arguments, and if it fails, log the given exception_message.
    If logging.DEBUG is set for the logger, then we also log the traceback.

    :param func: function to call
    :param args: arguments to pass
    :param exception_message: message to log on exception
    :return: whatever the func returns
    """
    try:
        return func(*args)
    except Exception as e:
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.exception(exception_message)
        else:
            LOG.error("%s: %s", exception_message, e)


class PluginLifecycleNotifierMixin:
    """
    Mixin that provides functions to dispatch calls to a PluginLifecycleListener in a safe way.
    """

    listeners: t.List[PluginLifecycleListener]

    def _fire_on_resolve_after(self, plugin_spec):
        for listener in self.listeners:
            _call_safe(
                listener.on_resolve_after,
                (plugin_spec,),  #
                "error while calling on_resolve_after",
            )

    def _fire_on_resolve_exception(self, namespace, entrypoint, exception):
        for listener in self.listeners:
            _call_safe(
                listener.on_resolve_exception,
                (namespace, entrypoint, exception),
                "error while calling on_resolve_exception",
            )

    def _fire_on_init_after(self, plugin_spec, plugin):
        for listener in self.listeners:
            _call_safe(
                listener.on_init_after,
                (
                    plugin_spec,
                    plugin,
                ),  #
                "error while calling on_init_after",
            )

    def _fire_on_init_exception(self, plugin_spec, exception):
        for listener in self.listeners:
            _call_safe(
                listener.on_init_exception,
                (plugin_spec, exception),
                "error while calling on_init_exception",
            )

    def _fire_on_load_before(self, plugin_spec, plugin, load_args, load_kwargs):
        for listener in self.listeners:
            _call_safe(
                listener.on_load_before,
                (plugin_spec, plugin, load_args, load_kwargs),
                "error while calling on_load_before",
            )

    def _fire_on_load_after(self, plugin_spec, plugin, result):
        for listener in self.listeners:
            _call_safe(
                listener.on_load_after,
                (plugin_spec, plugin, result),
                "error while calling on_load_after",
            )

    def _fire_on_load_exception(self, plugin_spec, plugin, exception):
        for listener in self.listeners:
            _call_safe(
                listener.on_load_exception,
                (plugin_spec, plugin, exception),
                "error while calling on_load_exception",
            )


class PluginContainer(t.Generic[P]):
    """
    Object to pass around the plugin state inside a PluginManager.
    """

    name: str
    lock: threading.RLock

    plugin_spec: PluginSpec
    plugin: P = None
    load_value: t.Any = None

    is_init: bool = False
    is_loaded: bool = False

    init_error: Exception = None
    load_error: Exception = None

    @property
    def distribution(self) -> Distribution:
        """
        Uses metadata from importlib to resolve the distribution information for this plugin.

        :return: the importlib.metadata.Distribution object
        """
        return resolve_distribution_information(self.plugin_spec)


class PluginManager(PluginLifecycleNotifierMixin, t.Generic[P]):
    """
    Manages Plugins within a namespace discovered by a PluginFinder. The default mechanism is to resolve plugins from
    entry points using a StevedorePluginFinder.

    A Plugin that is managed by a PluginManager can be in three states:
        * resolved: the entrypoint pointing to the PluginSpec was imported and the PluginSpec instance was created
        * init: the PluginFactory of the PluginSpec was successfully invoked
        * loaded: the load method of the Plugin was successfully invoked

    Internally, the PluginManager uses PluginContainer instances to keep the state of Plugin instances.
    """

    namespace: str

    load_args: t.Union[t.List, t.Tuple]
    load_kwargs: t.Dict[str, t.Any]
    listeners: t.List[PluginLifecycleListener]

    def __init__(
        self,
        namespace: str,
        load_args: t.Union[t.List, t.Tuple] = None,
        load_kwargs: t.Dict = None,
        listener: t.Union[PluginLifecycleListener, t.Iterable[PluginLifecycleListener]] = None,
        finder: PluginFinder = None,
    ):
        self.namespace = namespace

        self.load_args = load_args or list()
        self.load_kwargs = load_kwargs or dict()

        if listener:
            if isinstance(listener, (list, set, tuple)):
                self.listeners = list(listener)
            else:
                self.listeners = [listener]
        else:
            self.listeners = []

        self.finder = finder or StevedorePluginFinder(
            self.namespace, self._fire_on_resolve_exception
        )

        self._plugin_index = None
        self._init_mutex = threading.RLock()

    def add_listener(self, listener: PluginLifecycleListener):
        self.listeners.append(listener)

    def load(self, name: str) -> P:
        """
        Loads the Plugin with the given name using the load args and kwargs set in the plugin manager constructor.
        If at any point in the lifecycle the plugin loading fails, the load method will raise the respective exception.

        Load is idempotent, so once the plugin is loaded, load will return the same instance again.
        """
        container = self._require_plugin(name)

        if not container.is_loaded:
            self._load_plugin(container)

        if container.init_error:
            raise container.init_error

        if container.load_error:
            raise container.load_error

        if not container.is_loaded:
            raise PluginException(
                "plugin did not load correctly", namespace=self.namespace, name=name
            )

        return container.plugin

    def load_all(self, propagate_exceptions=False) -> t.List[P]:
        """
        Attempts to load all plugins found in the namespace, and returns those that were loaded successfully. If
        propagate_exception is set to True, then the method will re-raise any errors as soon as it encouters them.
        """
        plugins = list()

        for name, container in self._plugins.items():
            if container.is_loaded:
                plugins.append(container.plugin)
                continue

            try:
                plugin = self.load(name)
                plugins.append(plugin)
            except PluginDisabled as e:
                LOG.debug("%s", e)
            except Exception as e:
                if propagate_exceptions:
                    raise
                else:
                    LOG.error("exception while loading plugin %s:%s: %s", self.namespace, name, e)

        return plugins

    def list_plugin_specs(self) -> t.List[PluginSpec]:
        return [container.plugin_spec for container in self._plugins.values()]

    def list_names(self) -> t.List[str]:
        return [spec.name for spec in self.list_plugin_specs()]

    def list_containers(self) -> t.List[PluginContainer[P]]:
        return list(self._plugins.values())

    def get_container(self, name: str) -> PluginContainer[P]:
        return self._require_plugin(name)

    def exists(self, name: str) -> bool:
        return name in self._plugins

    def is_loaded(self, name: str) -> bool:
        return self._require_plugin(name).is_loaded

    @property
    def _plugins(self) -> t.Dict[str, PluginContainer[P]]:
        if self._plugin_index is None:
            with self._init_mutex:
                if self._plugin_index is None:
                    self._plugin_index = self._init_plugin_index()

        return self._plugin_index

    def _require_plugin(self, name: str) -> PluginContainer[P]:
        if name not in self._plugins:
            raise ValueError("no plugin named %s in namespace %s" % (name, self.namespace))

        return self._plugins[name]

    def _load_plugin(self, container: PluginContainer):
        with container.lock:
            plugin_spec = container.plugin_spec

            # instantiate Plugin from spec if necessary
            if not container.is_init:
                try:
                    LOG.debug("instantiating plugin %s", plugin_spec)
                    container.plugin = self._plugin_from_spec(plugin_spec)
                    container.is_init = True
                    self._fire_on_init_after(plugin_spec, container.plugin)
                except Exception as e:
                    # TODO: maybe we should move these logging blocks to `load_all`, since this is the only instance
                    #  where exceptions messages may get lost.
                    if LOG.isEnabledFor(logging.DEBUG):
                        LOG.exception("error instantiating plugin %s", plugin_spec)

                    self._fire_on_init_exception(plugin_spec, e)
                    container.init_error = e
                    return

            plugin = container.plugin

            if not plugin.should_load():
                raise PluginDisabled(namespace=self.namespace, name=container.plugin_spec.name)

            args = self.load_args
            kwargs = self.load_kwargs

            self._fire_on_load_before(plugin_spec, plugin, args, kwargs)
            try:
                LOG.debug("loading plugin %s:%s", self.namespace, plugin_spec.name)
                result = plugin.load(*args, **kwargs)
                self._fire_on_load_after(plugin_spec, plugin, result)
                container.load_value = result
                container.is_loaded = True
            except Exception as e:
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.exception("error loading plugin %s", plugin_spec)
                self._fire_on_load_exception(plugin_spec, plugin, e)
                container.load_error = e

    def _plugin_from_spec(self, plugin_spec: PluginSpec) -> P:
        factory = plugin_spec.factory

        # functional decorators can overwrite the spec factory (pointing to the decorator) with a custom factory
        spec = getattr(factory, "__pluginspec__", None)
        if spec:
            factory = spec.factory

        return factory()

    def _init_plugin_index(self) -> t.Dict[str, PluginContainer]:
        return {plugin.name: plugin for plugin in self._import_plugins() if plugin}

    def _import_plugins(self) -> t.Iterable[PluginContainer]:
        for spec in self.finder.find_plugins():
            self._fire_on_resolve_after(spec)

            if spec.namespace != self.namespace:
                continue

            yield self._create_container(spec)

    def _create_container(self, plugin_spec: PluginSpec) -> PluginContainer:
        container = PluginContainer()
        container.lock = threading.RLock()
        container.name = plugin_spec.name
        container.plugin_spec = plugin_spec
        return container
